﻿using Nebulix;
using Nebulix.Rendering;
using Silk.NET.Maths;
using Silk.NET.OpenGL;

namespace Engine_silk.NET
{
	// also maybe make an ISphere and call this class "CubeSphere"?
	public class Sphere
	{
		private readonly Face[] sphereFaces = new Face[6];
		private readonly uint resolution;
		
		public Sphere(uint resolution)
		{
			if (resolution < 2)
				throw new ArgumentOutOfRangeException(nameof(resolution), resolution,
					"Resolution must be greater than 1");
			this.resolution = resolution;
		}

		public void CreateSphere()
		{
			Vector3D<float>[] directions =
			[
				Vector3D<float>.UnitZ, -Vector3D<float>.UnitZ, 
				Vector3D<float>.UnitY, -Vector3D<float>.UnitY,
				Vector3D<float>.UnitX, -Vector3D<float>.UnitX
			];

			for (int i = 0; i < sphereFaces.Length; i++)
			{
				sphereFaces[i] = new Face(directions[i], new Mesh(), resolution);
				sphereFaces[i].GenerateMesh();
			}
		}
		
		/// <summary>
		/// Renders the mesh. No Shader is specified and <see cref="CreateSphere"/> needs to be called beforehand 
		/// </summary>
		public void RenderSphere(GL gl) // Will not be needed  
		{
			// Use default shader etc. to render the sphere
			for (int i = 0; i < 6; i++)
			{
				sphereFaces[i].Mesh.Render(gl);
			}
		}
	}

	internal record struct Face
	{
		public Mesh Mesh => _mesh;

		private readonly Mesh _mesh;
		private readonly Vector3D<float> _localX;
		private readonly Vector3D<float> _localY;

		private readonly Vector3D<float> _localUp;
		private readonly uint _resolution;
		
		public Face(Vector3D<float> localUp, Mesh mesh, uint resolution)
		{
			_mesh = mesh;
			_localX = new(localUp.Y, localUp.Z, localUp.X);
			_localY = Vector3D.Cross(localUp, _localX);
			_localUp = localUp;
			_resolution = resolution;
		}

		internal void GenerateMesh()
		{
			var vertices = new Vector3D<float>[_resolution * _resolution];
			// _resolution - 1 because the vertices index starts at 0 
			// * 6 because each triangle needs 3 points and each small quad has 2 triangles 3*2 = 6
			var indices = new uint[(_resolution - 1) * (_resolution - 1) * 6];
			int triangleIndex = 0;

			uint i;
			for (uint y = 0; y < _resolution; y++)
			{
				for (uint x = 0; x < _resolution; x++)
				{
					i = x + y * _resolution;
					Vector2D<float> percent = new Vector2D<float>(x, y) / (_resolution - 1);
					// place vertex on correct position of the plane to easily calculate indices
					Vector3D<float> vertexPosition = _localUp + (percent.X - 0.5f) * 2 * _localX + (percent.Y - 0.5f) * 2 * _localY;
					vertices[i] = PointOnCubeToPointOnSphere(vertexPosition);

					if (x != _resolution - 1 && y != _resolution - 1) // we didn't reach the bottom right point yet
					{
						indices[triangleIndex] = i;
						indices[triangleIndex + 1] = i + _resolution + 1;
						indices[triangleIndex + 2] = i + _resolution;
						
						indices[triangleIndex + 3] = i;
						indices[triangleIndex + 4] = i + 1;
						indices[triangleIndex + 5] = i + _resolution + 1;

						triangleIndex += 6; // doing it this way is faster than doing indices[triangleIndex++] (takes about double the time)
					}
				}
			}

			_mesh.Clear();
			_mesh.Vertices = vertices;
			_mesh.Indices = indices;
			_mesh.CalculateNormals();
		}

		// Smooth mapping so that the points are not clumped on the former corners of the cube
		// http://mathproofs.blogspot.com/2005/07/mapping-cube-to-sphere.html
		private Vector3D<float> PointOnCubeToPointOnSphere(Vector3D<float> point)
		{
			float x2 = point.X * point.X;
			float y2 = point.Y * point.Y;
			float z2 = point.Z * point.Z;
			float newX = point.X * MathF.Sqrt(1 - y2 / 2 - z2 / 2 + (y2 * z2) / 3);
			float newY = point.Y * MathF.Sqrt(1 - z2 / 2 - x2 / 2 + (z2 * x2) / 3);
			float newZ = point.Z * MathF.Sqrt(1 - x2 / 2 - y2 / 2 + (x2 * y2) / 3);
			return new Vector3D<float>(newX, newY, newZ);
		}
	}
}
